/*
 *  AMT IDE-R functions.
 *
 *  Copyright (C) 2014 Andreas Steinmetz <ast@domdv.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <poll.h>
#include <unistd.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "auth.h"
#include "ider.h"

/* known message types */

#define OPEN_SESSION	0
#define CLOSE_SESSION	1
#define PING_PONG	2
#define STATUS_NOTIFY	3
#define STATUS_DATA	4
#define ABORT		5
#define HEARTBEAT	6
#define SCSI_CMD	7
#define SCSI_SENSE	8
#define SCSI_READ	9
#define SCSI_CHUNK	10
#define SCSI_WRITE	11

/* STATUS_DATA values */

#define REGS_AVAIL	1
#define REGS_STATUS	2
#define REGS_TOGGLE	3

/* status notify values */

#define RESET_NOTIFY	3

/* REGS_AVAIL bits on receive */

#define BIT_PRI_AVAIL	0x01
#define BIT_SEC_AVAIL	0x02

/* REGS_STATUS bits on receive */

#define BIT_PRI_DFLT	0x01
#define BIT_PRI_CURR	0x02
#define BIT_SEC_DFLT	0x04
#define BIT_SEC_CURR	0x08

/* REGS_TOGGLE bits on send */

#define BIT_PRI_ENABLE	0x01
#define BIT_PRI_DISABLE	0x02
#define BITS_PRI_REBOOT	0x08
#define BITS_PRI_GRACE	0x10
#define BITS_PRI_NOW	0x18

/* REGS_TOGGLE values on receive */

#define REGS_SUCCESS	0x01

/* known used scsi commands */

#define TEST_UNIT_READY			0x00
#define READ_6				0x08
#define WRITE_6				0x0a
#define MODE_SELECT_6			0x15
#define MODE_SENSE_6			0x1a
#define ALLOW_MEDIUM_REMOVAL		0x1e
#define READ_CAPACITY			0x25
#define READ_10				0x28
#define WRITE_10			0x2a
#define READ_TOC			0x43
#define GET_CONFIGURATION		0x46
#define GET_EVENT_STATUS_NOTIFICATION	0x4a
#define READ_DISC_INFO			0x51
#define MODE_SELECT_10			0x55
#define MODE_SENSE_10			0x5a

struct devinfo
{
	int fd;
	uint32_t blocks;
};

struct iderctx
{
	int fd;
	int persist;
	int verbose;
	int debug;
	int scsistate;
	int readbfr;
	int writebfr;
	int session;
	int rx_timeout;
	int tx_timeout;
	int heartbeat;
	int readtoc;
	int regs;
	int enabled;
	uint32_t pos;
	uint32_t len;
	uint32_t sndseq;
	uint32_t rcvseq;
	int (*io)(void *parm,unsigned char *data,int len,int mode);
	int (*iowait)(void *parm,int msecs);
	void (*regstate)(void *parm,int state);
	void *parm;
	struct devinfo dev[2];
};

struct redirinfo
{
	int major;
	int minor;
	int fwmajor;
	int fwminor;
	int iana;
	int length;
	struct in_addr addr;
	unsigned char data[255];
};

struct sessiondata
{
	int major;
	int minor;
	int fwmajor;
	int fwminor;
	int readbfr;	/* 3072 or 8192 */
	int writebfr;	/* 3072 or 8192 */
	int proto;	/* must be zero	*/
	int iana;
	int length;
	unsigned char data[255];
};

struct statusdata
{
	int type;
	int value;
};

struct notifydata
{
	int value;
};

struct scsidata
{
	int prisec;	/* always 0 as secondary testing is not possible */
	int device;
	unsigned char cdb[12];	/* a real scsi cdb of 6 10 or 12 bytes */
};

struct writedata
{
	int length;
	unsigned char data[8192];
};

struct abortdata
{
	int reason;
};

union msgdata
{
	struct sessiondata session;
	struct statusdata status;
	struct notifydata notify;
	struct scsidata scsi;
	struct writedata write;
	struct abortdata abort;
};

struct sessionparam
{
	int rx_timeout;
	int tx_timeout;
	int heartbeat;
	int version;	/* can be 0 or 1, changes readbfr and writebfr size */
};

struct dataparam
{
	int type;
	int value;
};

struct senseparam
{
	int prisec;	/* ignored as secondary testing is not possible */
	int device;
	int error;	/* 0 for ok, sense/asc/asq ignored, 1 for error */
	int sense;
	int asc;
	int asq;
};

struct readparam
{
	int complete;
	int prisec;	/* ignored as secondary testing is not possible */
	int device;
	int length;
	unsigned char data[8192];
};

struct chunkparam
{
	int prisec;	/* ignored as secondary testing is not possible */
	int device;
	int length;
};

union reqdata
{
	struct sessionparam session;
	struct dataparam data;
	struct senseparam sense;
	struct readparam read;
	struct chunkparam chunk;
};

static int start_redirect(struct iderctx *ctx,char *what,struct redirinfo *info)
{
	int length;
	unsigned char bfr[255];

	if(info)memset(info,0,sizeof(struct redirinfo));

	memset(bfr,0,sizeof(bfr));
	bfr[0]=0x10;
	if(ctx->persist)bfr[1]=0x01;
	memcpy(bfr+4,what,4);
	if(ctx->io(ctx->parm,bfr,8,WRITE))return -1;

	if(ctx->io(ctx->parm,bfr,13,READ))return -1;
	if(bfr[0]!=0x11)return -1;
	switch(bfr[1])
	{
	case 0x00:
		length=bfr[12];
		if(info)
		{
			info->major=bfr[4];
			info->minor=bfr[5];
			info->fwmajor=bfr[6];
			info->fwminor=bfr[7];
			info->iana=bfr[11];
			info->iana<<=8;
			info->iana|=bfr[10];
			info->iana<<=8;
			info->iana|=bfr[9];
			info->iana<<=8;
			info->iana|=bfr[8];
			info->length=length;
			if(ctx->io(ctx->parm,info->data,length,READ))return -1;
		}
		else if(ctx->io(ctx->parm,bfr,length,READ))return -1;
		return 0;

	case 0x02:
		if(info)memcpy(&info->addr,bfr+4,4);
		return -2;

	default:return -1;
	}

}

static int end_redirect(struct iderctx *ctx)
{
	unsigned char bfr[4];

	memset(bfr,0,sizeof(bfr));
	bfr[0]=0x12;
	if(ctx->io(ctx->parm,bfr,4,WRITE))return -1;

	return 0;
}

static int ider_recv(struct iderctx *ctx,int *msg,union msgdata *data)
{
	uint32_t seq;
	unsigned char hdr[8];
	unsigned char dta[22];

	if(ctx->io(ctx->parm,hdr,8,READ))
	{
		if(ctx->verbose)fprintf(stderr,"network read error\n");
		return -1;
	}

	switch(hdr[0])
	{
	/* open session ack */

	case 0x41:
		*msg=OPEN_SESSION;
		if(ctx->io(ctx->parm,dta,22,READ))
		{
			if(ctx->verbose)fprintf(stderr,"network read error\n");
			return -1;
		}
		data->session.major=dta[0];
		data->session.minor=dta[1];
		data->session.fwmajor=dta[2];
		data->session.fwminor=dta[3];
		data->session.readbfr=dta[9];
		data->session.readbfr<<=8;
		data->session.readbfr|=dta[8];
		data->session.writebfr=dta[11];
		data->session.writebfr<<=8;
		data->session.writebfr|=dta[10];
		data->session.proto=dta[13];
		data->session.iana=dta[20];
		data->session.iana<<=8;
		data->session.iana|=dta[19];
		data->session.iana<<=8;
		data->session.iana|=dta[18];
		data->session.iana<<=8;
		data->session.iana|=dta[17];
		data->session.length=dta[21];
		if(ctx->io(ctx->parm,data->session.data,data->session.length,
			READ))
		{
			if(ctx->verbose)fprintf(stderr,"network read error\n");
			return -1;
		}
		break;

	/* close session ack */

	case 0x43:
		*msg=CLOSE_SESSION;
		break;

	/* pong */

	case 0x45:
		*msg=PING_PONG;
		break;

	/* status notification */

	case 0x46:
		*msg=STATUS_NOTIFY;
		if(ctx->io(ctx->parm,dta,1,READ))
		{
			if(ctx->verbose)fprintf(stderr,"network read error\n");
			return -1;
		}
		data->notify.value=dta[0];
		break;

	/* status data */

	case 0x49:
		*msg=STATUS_DATA;
		if(ctx->io(ctx->parm,dta,5,READ))
		{
			if(ctx->verbose)fprintf(stderr,"network read error\n");
			return -1;
		}
		data->status.type=dta[0];
		data->status.value=dta[4];
		data->status.value<<=8;
		data->status.value|=dta[3];
		data->status.value<<=8;
		data->status.value|=dta[2];
		data->status.value<<=8;
		data->status.value|=dta[1];
		break;

	/* probably cancel/reset scsi command */

	case 0x4a:
		*msg=ABORT;
		if(ctx->io(ctx->parm,dta,3,READ))
		{
			if(ctx->verbose)fprintf(stderr,"network read error\n");
			return -1;
		}
		data->abort.reason=dta[0];
		break;

	/* heartbeat */

	case 0x4b:
		*msg=HEARTBEAT;
		break;

	/* scsi command */

	case 0x50:
		*msg=SCSI_CMD;
		if(ctx->io(ctx->parm,dta,20,READ))
		{
			if(ctx->verbose)fprintf(stderr,"network read error\n");
			return -1;
		}
		data->scsi.prisec=0;
		data->scsi.device=dta[6]&0xbf;
		memcpy(data->scsi.cdb,dta+8,12);
		break;

	/* scsi write data */

	case 0x53:
		*msg=SCSI_WRITE;
		if(ctx->io(ctx->parm,dta,6,READ))
		{
			if(ctx->verbose)fprintf(stderr,"network read error\n");
			return -1;
		}
		data->write.length=dta[2];
		data->write.length<<=8;
		data->write.length|=dta[1];
		if(ctx->io(ctx->parm,data->write.data,data->write.length,READ))
		{
			if(ctx->verbose)fprintf(stderr,"network read error\n");
			return -1;
		}
		break;

	/* unknown */

	default:if(ctx->verbose)fprintf(stderr,"unknown message 0x%02x "
			"received\n",hdr[0]);
		return -1;
	}

	if(hdr[1]||hdr[2]||hdr[3])if(ctx->verbose)fprintf(stderr,
		"reserved bytes are %02x %02x %02x, should be zero\n",
		hdr[1],hdr[2],hdr[3]);

	seq=hdr[7];
	seq<<=8;
	seq|=hdr[6];
	seq<<=8;
	seq|=hdr[5];
	seq<<=8;
	seq|=hdr[4];

	if(seq!=++ctx->rcvseq)
	{
		if(ctx->verbose)fprintf(stderr,"sequence number error, is "
			"%08x, should be %08x\n",seq,ctx->rcvseq);
		return -1;
	}

	return 0;
}

static int ider_send(struct iderctx *ctx,int msg,union reqdata *data)
{
	int len=8;
	int complete=1;
	unsigned char bfr[8448];

	switch(msg)
	{
	case OPEN_SESSION:
		len=18;
		bfr[0]=0x40;
		bfr[8]=data->session.rx_timeout&0xff;
		bfr[9]=(unsigned char)(data->session.rx_timeout>>8);
		bfr[10]=data->session.tx_timeout&0xff;
		bfr[11]=(unsigned char)(data->session.tx_timeout>>8);
		bfr[12]=data->session.heartbeat&0xff;
		bfr[13]=(unsigned char)(data->session.heartbeat>>8);
		bfr[14]=data->session.version&0xff;
		bfr[15]=(unsigned char)(data->session.version>>8);
		bfr[16]=(unsigned char)(data->session.version>>16);
		bfr[17]=(unsigned char)(data->session.version>>24);
		break;

	case CLOSE_SESSION:
		bfr[0]=0x42;
		break;

	case PING_PONG:
		bfr[0]=0x44;
		break;

	case STATUS_NOTIFY:
		bfr[0]=0x47;
		break;

	case STATUS_DATA:
		len=9;
		bfr[0]=0x48;
		if((bfr[8]=data->data.type&0xff)==REGS_TOGGLE)
		{
			len=13;
			bfr[9]=data->data.value&0xff;
			bfr[10]=(unsigned char)(data->data.value>>8);
			bfr[11]=(unsigned char)(data->data.value>>16);
			bfr[12]=(unsigned char)(data->data.value>>24);
		}
		break;

	case HEARTBEAT:
		bfr[0]=0x4b;
		break;

	case SCSI_SENSE:
		len=31;
		bfr[0]=0x51;
		memset(bfr+8,0,23);
		if(!data->sense.error)bfr[20]=0xc5;
		else
		{
			bfr[20]=0x87;
			bfr[21]=data->sense.sense<<4;
		}
		bfr[22]=0x03;
		bfr[26]=data->sense.device&0xff;
		if(!data->sense.error)bfr[27]=0x50;
		else
		{
			bfr[27]=0x51;
			bfr[28]=data->sense.sense;
			bfr[29]=data->sense.asc;
			bfr[30]=data->sense.asq;
		}
		break;

	case SCSI_READ:
		len=34+data->read.length;
		bfr[0]=0x54;
		memset(bfr+8,0,26);
		bfr[9]=bfr[16]=data->read.length&0xff;
		bfr[10]=bfr[17]=(unsigned char)(data->read.length>>8);
		bfr[12]=0xb5;
		bfr[14]=0x02;
		bfr[18]=data->read.device&0xff;
		bfr[19]=0x58;
		if(data->read.complete)
		{
			bfr[20]=0x85;
			bfr[22]=0x03;
			bfr[26]=data->read.device&0xff;
			bfr[27]=0x50;
		}
		else complete=0;
		memcpy(bfr+34,data->read.data,data->read.length);
		break;

	case SCSI_CHUNK:
		complete=0;
		len=31;
		bfr[0]=0x52;
		memset(bfr+8,0,23);
		bfr[9]=bfr[16]=data->chunk.length&0xff;
		bfr[10]=bfr[17]=(unsigned char)(data->chunk.length>>8);
		bfr[12]=0xb5;
		bfr[18]=data->chunk.device&0xff;
		bfr[19]=0x58;
		break;

	default:if(ctx->verbose)fprintf(stderr,"parameter error\n");
		return -1;
	}

	bfr[1]=bfr[2]=bfr[3]=0;
	if(bfr[0]>=0x50&&complete)bfr[3]=0x02;

	ctx->sndseq++;
	bfr[4]=ctx->sndseq&0xff;
	bfr[5]=(unsigned char)(ctx->sndseq>>8);
	bfr[6]=(unsigned char)(ctx->sndseq>>16);
	bfr[7]=(unsigned char)(ctx->sndseq>>24);

	if(ctx->io(ctx->parm,bfr,len,WRITE))
	{
		if(ctx->verbose)fprintf(stderr,"network write error\n");
		return -1;
	}

	return 0;
}

static int handle_scsi(struct iderctx *ctx,int dev,unsigned char *cdb,
	union reqdata *data)
{
	uint32_t lba;
	uint32_t len;
	struct devinfo *di;

	switch(cdb[0])
	{
	case WRITE_10:
		lba=cdb[2];
		lba<<=8;
		lba|=cdb[3];
		lba<<=8;
		lba|=cdb[4];
		lba<<=8;
		lba|=cdb[5];
		len=cdb[7];
		len<<=8;
		len|=cdb[8];
		if(ctx->debug)fprintf(stderr,"WRITE_10 %02x %d %d\n",
			dev,lba,len);
		goto wrcmn;

	case WRITE_6:
		lba=cdb[1]&0x1f;
		lba<<=8;
		lba|=cdb[2];
		lba<<=8;
		lba|=cdb[3];
		len=cdb[4];
		if(!len)len=256;
		if(ctx->debug)fprintf(stderr,"WRITE_6 %02x %d %d\n",
			dev,lba,len);

wrcmn:		switch(dev)
		{
		case DEV_FLOPPY:
			if(!(ctx->scsistate&4))
			{
nomedium:			data->sense.device=dev;
				data->sense.error=1;
				data->sense.sense=0x02;
				data->sense.asc=0x3a;
				data->sense.asq=0x00;
				if(ider_send(ctx,SCSI_SENSE,data))return -1;
				return 0;
			}
			di=&ctx->dev[0];
			break;

		case DEV_CDDVD:
			if(!(ctx->scsistate&8))goto nomedium;
			di=&ctx->dev[1];
			break;

		default:if(ctx->verbose)fprintf(stderr,"internal error 1 "
				"%02x\n",dev);
			return -1;
		}

		if(lba+len<lba||lba+len>di->blocks)
		{
			data->sense.device=dev;
			data->sense.error=1;
			data->sense.sense=0x05;
			data->sense.asc=0x21;
			data->sense.asq=0x00;
			if(ider_send(ctx,SCSI_SENSE,data))return -1;
			return 0;
		}

		if(!len)
		{
			data->sense.device=dev;
			data->sense.error=0;
			if(ider_send(ctx,SCSI_SENSE,data))return -1;
			return 0;
		}

		switch(dev)
		{
		case DEV_FLOPPY:
			if(!(ctx->scsistate&16))break;
		case DEV_CDDVD:
			data->sense.device=dev;
			data->sense.error=1;
			data->sense.sense=0x07;
			data->sense.asc=0x27;
			data->sense.asq=0x00;
			if(ider_send(ctx,SCSI_SENSE,data))return -1;
			return 0;
		}

		ctx->pos=lba<<9;
		ctx->len=len<<9;

		data->chunk.device=dev;
		if(ctx->len>ctx->writebfr)data->chunk.length=ctx->writebfr;
		else data->chunk.length=ctx->len;
		if(ider_send(ctx,SCSI_CHUNK,data))return -1;
		break;

	case READ_10:
		lba=cdb[2];
		lba<<=8;
		lba|=cdb[3];
		lba<<=8;
		lba|=cdb[4];
		lba<<=8;
		lba|=cdb[5];
		len=cdb[7];
		len<<=8;
		len|=cdb[8];
		if(ctx->debug)fprintf(stderr,"READ_10 %02x %d %d\n",
			dev,lba,len);
		goto rdcmn;

	case READ_6:
		lba=cdb[1]&0x1f;
		lba<<=8;
		lba|=cdb[2];
		lba<<=8;
		lba|=cdb[3];
		len=cdb[4];
		if(!len)len=256;
		if(ctx->debug)fprintf(stderr,"READ_6 %02x %d %d\n",dev,lba,len);

rdcmn:		switch(dev)
		{
		case DEV_FLOPPY:
			if(!(ctx->scsistate&4))goto nomedium;
			di=&ctx->dev[0];
			break;

		case DEV_CDDVD:
			if(!(ctx->scsistate&8))goto nomedium;
			di=&ctx->dev[1];
			break;

		default:if(ctx->verbose)fprintf(stderr,"internal error 2 "
				"%02x\n",dev);
			return -1;
		}

		if(lba+len<lba||lba+len>di->blocks)
		{
			data->sense.device=dev;
			data->sense.error=1;
			data->sense.sense=0x05;
			data->sense.asc=0x21;
			data->sense.asq=0x00;
			if(ider_send(ctx,SCSI_SENSE,data))return -1;
			return 0;
		}

		if(!len)
		{
			data->sense.device=dev;
			data->sense.error=0;
			if(ider_send(ctx,SCSI_SENSE,data))return -1;
			return 0;
		}

		switch(dev)
		{
		case DEV_FLOPPY:
			lba<<=9;
			len<<=9;
			break;

		case DEV_CDDVD:
			lba<<=11;
			len<<=11;
			break;
		}

		if(lseek(di->fd,lba,SEEK_SET)==(off_t)-1)
		{
			if(ctx->verbose)perror("lseek");
			data->sense.device=dev;
			data->sense.error=1;
			data->sense.sense=0x03;
			data->sense.asc=0x02;
			data->sense.asq=0x00;
			if(ider_send(ctx,SCSI_SENSE,data))return -1;
			return 0;
		}

		data->read.device=dev;
		for(;len;len-=data->read.length)
		{
			if(len>ctx->readbfr)
			{
				data->read.complete=0;
				data->read.length=ctx->readbfr;
			}
			else
			{
				data->read.complete=1;
				data->read.length=len;
			}
			if(read(di->fd,data->read.data,data->read.length)!=
				data->read.length)
			{
				if(ctx->verbose)perror("read");
				data->sense.device=dev;
				data->sense.error=1;
				data->sense.sense=0x03;
				data->sense.asc=0x11;
				data->sense.asq=0x00;
				if(ider_send(ctx,SCSI_SENSE,data))return -1;
				return 0;
			}
			if(ider_send(ctx,SCSI_READ,data))return -1;
		}
		break;

	case TEST_UNIT_READY:
		if(ctx->debug)fprintf(stderr,"TEST_UNIT_READY %02x\n",dev);

		switch(dev)
		{
		case DEV_FLOPPY:
			if((ctx->scsistate&4)&&!(ctx->scsistate&1))
			{
				ctx->scsistate|=1;
				goto attrdy;
			}
			else if(!(ctx->scsistate&4))goto nomedium;
			break;

		case DEV_CDDVD:
			if((ctx->scsistate&8)&&!(ctx->scsistate&2))
			{
				ctx->scsistate|=2;
attrdy:				data->sense.device=dev;
				data->sense.error=1;
				data->sense.sense=0x06;
				data->sense.asc=0x28;
				data->sense.asq=0x00;
			}
			else if(!(ctx->scsistate&8))goto nomedium;
			break;

		default:if(ctx->verbose)fprintf(stderr,"internal error 3 "
				"%02x\n",dev);
			return -1;
		}

		data->sense.device=dev;
		data->sense.error=0;
		if(ider_send(ctx,SCSI_SENSE,data))return -1;
		break;

	case READ_CAPACITY:
		if(ctx->debug)fprintf(stderr,"READ_CAPACITY %02x\n",dev);
		data->read.device=dev;
		data->read.length=8;
		data->read.complete=1;
		switch(dev)
		{
		case DEV_FLOPPY:
			if(ctx->scsistate&4)len=ctx->dev[0].blocks-1;
			else goto nomedium;
			break;

		case DEV_CDDVD:
			if(ctx->scsistate&8)len=ctx->dev[1].blocks-1;
			else goto nomedium;
			break;

		default:if(ctx->verbose)fprintf(stderr,"internal error 4 "
				"%02x\n",dev);
			return -1;
		}

		data->read.data[0]=(unsigned char)(len>>24);
		data->read.data[1]=(unsigned char)(len>>16);
		data->read.data[2]=(unsigned char)(len>>8);
		data->read.data[3]=len&0xff;
		data->read.data[4]=0x00;
		data->read.data[5]=0x00;
		data->read.data[6]=(dev==DEV_CDDVD?0x08:0x02);
		data->read.data[7]=0x00;
		if(ider_send(ctx,SCSI_READ,data))return -1;
		break;

	case GET_EVENT_STATUS_NOTIFICATION:
		if(ctx->debug)fprintf(stderr,"GET_EVENT_STATUS_NOTIFICATION "
			"%02x %02x %02x %02x\n",dev,cdb[1],cdb[4],cdb[9]);
		if(cdb[1]!=0x01||cdb[4]!=0x10)
		{
invparm:		data->sense.device=dev;
			data->sense.error=1;
			data->sense.sense=0x05;
			data->sense.asc=0x26;
			data->sense.asq=0x01;
			if(ider_send(ctx,SCSI_SENSE,data))return -1;
			break;
		}
		data->read.device=dev;
		data->read.length=4;
		data->read.complete=1;
		data->read.data[0]=0x00;
		data->read.data[1]=0x00;
		switch(dev)
		{
		case DEV_FLOPPY:
			if(ctx->scsistate&4)data->read.data[1]=0x02;
			break;

		case DEV_CDDVD:
			if(ctx->scsistate&8)data->read.data[1]=0x02;
			break;

		default:if(ctx->verbose)fprintf(stderr,"internal error 5 "
				"%02x\n",dev);
			return -1;
		}
		data->read.data[2]=0x80;
		data->read.data[3]=0x00;
		if(ider_send(ctx,SCSI_READ,data))return -1;
		break;

	case MODE_SENSE_6:
		if(ctx->debug)fprintf(stderr,"MODE_SENSE_10 %02x\n",dev);
		if(cdb[2]==0x3f&&cdb[3]==0x00)
		{
			data->read.device=dev;
			data->read.length=4;
			data->read.complete=1;
			data->read.data[0]=0x00;
			switch(dev)
			{
			case DEV_FLOPPY:
				if(!(ctx->scsistate&4))goto nomedium;
				data->read.data[1]=0x00;
				if(ctx->scsistate&16)data->read.data[2]=0x80;
				else data->read.data[2]=0x00;
				break;

			case DEV_CDDVD:
				if(!(ctx->scsistate&8))goto nomedium;
				data->read.data[1]=0x05;
				data->read.data[2]=0x80;
				break;

			default:if(ctx->verbose)
					fprintf(stderr,"internal error 6 "
						"%02x\n",dev);
				return -1;
			}
			data->read.data[3]=0x00;
			if(ider_send(ctx,SCSI_READ,data))return -1;
			break;
		}
		data->sense.device=dev;
		data->sense.error=1;
		data->sense.sense=0x05;
		data->sense.asc=0x24;
		data->sense.asq=0x00;
		if(ider_send(ctx,SCSI_SENSE,data))return -1;
		break;

	case MODE_SENSE_10:
		if(ctx->debug)fprintf(stderr,"MODE_SENSE_10 %02x\n",dev);
		if(cdb[2]==0x3f&&cdb[3]==0x00)
		{
			data->read.device=dev;
			data->read.length=8;
			data->read.complete=1;
			data->read.data[0]=0x00;
			data->read.data[1]=0x00;
			switch(dev)
			{
			case DEV_FLOPPY:
				if(!(ctx->scsistate&4))goto nomedium;
				data->read.data[2]=0x00;
				if(ctx->scsistate&16)data->read.data[3]=0x80;
				else data->read.data[3]=0x00;
				break;

			case DEV_CDDVD:
				if(!(ctx->scsistate&8))goto nomedium;
				data->read.data[2]=0x05;
				data->read.data[3]=0x80;
				break;

			default:if(ctx->verbose)
					fprintf(stderr,"internal error 7 "
						"%02x\n",dev);
				return -1;
			}
			data->read.data[4]=0x00;
			data->read.data[5]=0x00;
			data->read.data[6]=0x00;
			data->read.data[7]=0x00;
			if(ider_send(ctx,SCSI_READ,data))return -1;
			break;
		}
		data->sense.device=dev;
		data->sense.error=1;
		data->sense.sense=0x05;
		data->sense.asc=0x24;
		data->sense.asq=0x00;
		if(ider_send(ctx,SCSI_SENSE,data))return -1;
		break;

	case ALLOW_MEDIUM_REMOVAL:
		if(ctx->debug)fprintf(stderr,"ALLOW_MEDIUM_REMOVAL %02x\n",dev);
		switch(dev)
		{
		case DEV_FLOPPY:
			if(!(ctx->scsistate&4))goto nomedium;
			break;

		case DEV_CDDVD:
			if(!(ctx->scsistate&8))goto nomedium;
			break;

		default:if(ctx->verbose)fprintf(stderr,"internal error 8 "
				"%02x\n",dev);
			return -1;
		}
		data->sense.device=dev;
		data->sense.error=0;
		if(ider_send(ctx,SCSI_SENSE,data))return -1;
		break;

	case READ_TOC:
		if(ctx->debug)fprintf(stderr,"READ_TOC %02x %02x %02x %02x\n",
			dev,cdb[2],cdb[6],cdb[9]);
		switch(dev)
		{
		case DEV_FLOPPY:
			goto notimpl;
			break;

		case DEV_CDDVD:
			if(!ctx->readtoc)goto notimpl;
			if(!(ctx->scsistate&8))goto nomedium;
			break;

		default:if(ctx->verbose)fprintf(stderr,"internal error 9 "
				"%02x\n",dev);
			return -1;
		}

		switch(cdb[2])
		{
		case 0x00:
			if(cdb[6]||cdb[9])goto invparm;
			data->read.device=dev;
			data->read.length=12;
			data->read.complete=1;
			data->read.data[0]=0x00;
			data->read.data[1]=0x08;
			data->read.data[2]=0x00;
			data->read.data[3]=0x00;
			data->read.data[4]=0x00;
			data->read.data[5]=0x06;
			data->read.data[6]=0x00;
			data->read.data[7]=0x00;
			data->read.data[8]=0x00;
			data->read.data[9]=0x00;
			data->read.data[10]=0x00;
			data->read.data[11]=0x00;
			if(ider_send(ctx,SCSI_READ,data))return -1;
			break;

		default:goto invparm;
		}
		break;

	case READ_DISC_INFO:
		if(ctx->debug)fprintf(stderr,"READ_DISC_INFO %02x\n",dev);
		goto notimpl;

	case GET_CONFIGURATION:
		if(ctx->debug)fprintf(stderr,"GET_CONFIGURATION %02x\n",dev);
		goto notimpl;

	case MODE_SELECT_6:
		if(ctx->debug)fprintf(stderr,"MODE_SELECT_6 %02x\n",dev);
		goto notimpl;

	case MODE_SELECT_10:
		if(ctx->debug)fprintf(stderr,"MODE_SELECT_10 %02x\n",dev);
		goto notimpl;

	default:if(ctx->verbose)fprintf(stderr,"unknown scsi command 0x%02x\n",
			cdb[0]);
notimpl:	data->sense.device=dev;
		data->sense.error=1;
		data->sense.sense=0x05;
		data->sense.asc=0x20;
		data->sense.asq=0x00;
		if(ider_send(ctx,SCSI_SENSE,data))return -1;
		break;
	}

	return 0;
}

static int handle_write(struct iderctx *ctx,unsigned char *data,int length,
	union reqdata *dta)
{
	if(length>ctx->len)
	{
		if(ctx->verbose)fprintf(stderr,"write sequence error\n");
		return -1;
	}

	if(lseek(ctx->dev[0].fd,ctx->pos,SEEK_SET)==(off_t)-1)
	{
		if(ctx->verbose)perror("lseek");
		ctx->pos=0;
		ctx->len=0;
		dta->sense.device=DEV_FLOPPY;
		dta->sense.error=1;
		dta->sense.sense=0x03;
		dta->sense.asc=0x02;
		dta->sense.asq=0x00;
		if(ider_send(ctx,SCSI_SENSE,dta))return -1;
		return 0;
	}

	if(write(ctx->dev[0].fd,data,length)!=length)
	{
		if(ctx->verbose)perror("write");
		ctx->pos=0;
		ctx->len=0;
		dta->sense.device=DEV_FLOPPY;
		dta->sense.error=1;
		dta->sense.sense=0x03;
		dta->sense.asc=0x11;
		dta->sense.asq=0x00;
		if(ider_send(ctx,SCSI_SENSE,dta))return -1;
		return 0;
	}

	ctx->pos+=length;
	ctx->len-=length;

	if(ctx->len)
	{
		dta->chunk.device=DEV_FLOPPY;
		if(ctx->len>ctx->writebfr)dta->chunk.length=ctx->writebfr;
		else dta->chunk.length=ctx->len;
		if(ider_send(ctx,SCSI_CHUNK,dta))return -1;
	}
	else
	{
		ctx->pos=0;
		dta->sense.device=DEV_FLOPPY;
		dta->sense.error=0;
		if(ider_send(ctx,SCSI_SENSE,dta))return -1;
	}

	return 0;
}

int ider_engine(struct iderctx *ctx)
{
	int msg;
	int fail=0;
	int delay;
	union msgdata rcvdata;
	union reqdata snddata;

	if(!ctx->heartbeat||ctx->heartbeat>ctx->rx_timeout)
		delay=ctx->rx_timeout-2000;
	else delay=ctx->heartbeat;

repeat:	switch(ctx->iowait(ctx->parm,delay))
	{
	case 0:	if(++fail<2)
		{
			if(!ctx->heartbeat||ctx->heartbeat>ctx->rx_timeout)
				if(ider_send(ctx,PING_PONG,NULL))return -1;
			goto repeat;
		}
	case -1:return -1;
	}

	if(ider_recv(ctx,&msg,&rcvdata))return -1;

	switch(msg)
	{
	case OPEN_SESSION:
		if(ctx->debug)fprintf(stderr,"OPEN_SESSION\n");
		if(rcvdata.session.proto)
		{
			if(ctx->verbose)fprintf(stderr,"unknown proto %d\n",
				rcvdata.session.proto);
			return -1;
		}
		if(rcvdata.session.readbfr>8192)
		{
			if(ctx->verbose)fprintf(stderr,"illegal read buffer "
				"size %d\n",rcvdata.session.readbfr);
			return -1;
		}
		if(rcvdata.session.writebfr>8192)
		{
			if(ctx->verbose)fprintf(stderr,"illegal write buffer "
				"size %d\n",rcvdata.session.writebfr);
			return -1;
		}

		ctx->readbfr=rcvdata.session.readbfr;
		ctx->writebfr=rcvdata.session.writebfr;
		ctx->pos=0;
		ctx->len=0;
		ctx->session=1;

		snddata.data.type=REGS_AVAIL;
		if(ider_send(ctx,STATUS_DATA,&snddata))return -1;
		break;

	case CLOSE_SESSION:
		if(ctx->debug)fprintf(stderr,"CLOSE_SESSION\n");
		ctx->session=0;
		return 1;

	case PING_PONG:
		if(ctx->debug)fprintf(stderr,"PING_PONG\n");
		break;

	case STATUS_NOTIFY:
		if(ctx->debug)fprintf(stderr,"STATUS_NOTIFY\n");
		if(ider_send(ctx,STATUS_NOTIFY,NULL))return -1;
		if(rcvdata.notify.value==RESET_NOTIFY&&ctx->regs==1)
		{
			snddata.data.type=REGS_STATUS;
			if(ider_send(ctx,STATUS_DATA,&snddata))return -1;
		}
		break;

	case STATUS_DATA:
		if(ctx->debug)fprintf(stderr,"STATUS_DATA\n");
		switch(rcvdata.status.type)
		{
		case REGS_AVAIL:
			if(rcvdata.status.value&BIT_PRI_AVAIL)
			{
				ctx->regs=1;
				snddata.data.type=REGS_STATUS;
				if(ider_send(ctx,STATUS_DATA,&snddata))
					return -1;
			}
			else
			{
				ctx->regs=0;
				if(ctx->regstate)
					ctx->regstate(ctx->parm,REGS_NONE);
			}
			break;

		case REGS_STATUS:
			if(rcvdata.status.value&BIT_PRI_CURR)
			{
				ctx->enabled=1;
				if(ctx->regstate)
					ctx->regstate(ctx->parm,REGS_ENABLED);
			}
			else
			{
				ctx->enabled=0;
				if(ctx->regstate)
					ctx->regstate(ctx->parm,REGS_DISABLED);
			}
			break;

		case REGS_TOGGLE:
			if(rcvdata.status.value!=REGS_SUCCESS)
			{
				if(ctx->regstate)
					ctx->regstate(ctx->parm,REGS_FAIL);
				if(ctx->verbose)fprintf(stderr,"register "
					"toggle failure\n");
			}
			snddata.data.type=REGS_STATUS;
			if(ider_send(ctx,STATUS_DATA,&snddata))return -1;
			break;
		}
		break;

	case ABORT:
		if(ctx->debug)fprintf(stderr,"ABORT");
		ctx->pos=0;
		ctx->len=0;
		break;

	case HEARTBEAT:
		if(ctx->debug)fprintf(stderr,"HEARTBEAT\n");
		if(ider_send(ctx,HEARTBEAT,NULL))return -1;
		break;

	case SCSI_CMD:
		if(ctx->debug)fprintf(stderr,"SCSI_CMD\n");
		if(!ctx->session)break;
		if(handle_scsi(ctx,rcvdata.scsi.device,
			rcvdata.scsi.cdb,&snddata))return -1;
		break;

	case SCSI_WRITE:
		if(ctx->debug)fprintf(stderr,"SCSI_WRITE\n");
		if(!ctx->session)break;
		if(handle_write(ctx,rcvdata.write.data,
			rcvdata.write.length,&snddata))return -1;
		break;
	}

	return 0;
}

int ider_device(struct iderctx *ctx,int dev,int mode,int fd)
{
	int bits1;
	int bits2;
	int bits3;
	struct devinfo *di;
	struct stat stb;

	switch(dev)
	{
	case DEV_FLOPPY:
		switch(mode)
		{
		case DEV_NONE:
			if(ctx->scsistate&4)
			{
				ctx->scsistate&=~20;
				ctx->scsistate|=1;
			}
			return 0;

		case DEV_RW:
		case DEV_RO:
			if(ctx->scsistate&4)
			{
dupimg:				if(ctx->verbose)fprintf(stderr,
					"duplicate image attach\n");
				return -1;
			}
			di=&ctx->dev[0];
			bits1=4+(mode==DEV_RO?16:0);
			bits2=~1;
			bits3=9;
			break;

		default:if(ctx->verbose)fprintf(stderr,"illegal mode\n");
			return -1;
		}
		break;

	case DEV_CDDVD:
		switch(mode)
		{
		case DEV_NONE:
			if(ctx->scsistate&8)
			{
				ctx->scsistate&=~8;
				ctx->scsistate|=2;
			}
			return 0;;

		case DEV_RW:
			if(ctx->verbose)fprintf(stderr,"CD/DVD must be RO\n");
			return -1;

		case DEV_RO:
			if(ctx->scsistate&8)goto dupimg;
			di=&ctx->dev[1];
			bits1=8;
			bits2=~2;
			bits3=11;
			break;

		default:if(ctx->verbose)fprintf(stderr,"illegal mode\n");
			return -1;
		}
		break;

	default:if(ctx->verbose)fprintf(stderr,"illegal device\n");
		return -1;
	}

	if(fstat(fd,&stb))
	{
		if(ctx->verbose)perror("fstat");
		return -1;
	}
	if(!S_ISREG(stb.st_mode))
	{
		if(ctx->verbose)fprintf(stderr,"can attach only image files\n");
		return -1;
	}
	if(!stb.st_size||stb.st_size&((1<<bits3)-1))
	{
		if(ctx->verbose)fprintf(stderr,
			"can attach only multiples of 512/2048\n");
		return -1;
	}
	if(stb.st_size>0x80000000)
	{
		if(ctx->verbose)fprintf(stderr,
			"library can handle up to 2GB only\n");
		return -1;
	}

	di->fd=fd;
	di->blocks=stb.st_size>>bits3;
	ctx->scsistate&=bits2;
	ctx->scsistate|=bits1;
	return 0;
}

int ider_regs(struct iderctx *ctx,int mode)
{
	struct dataparam data;

	data.type=REGS_TOGGLE;

	switch(mode)
	{
	case REGS_STATE:
		switch(ctx->regs)
		{
		case 0:	return REGS_NONE;
		case 1:	switch(ctx->enabled)
			{
			case 0:	return REGS_DISABLED;
			case 1:	return REGS_ENABLED;
			}
		}
		return REGS_UNKNOWN;

	case REGS_ENABLE_NOW:
	case REGS_ENABLE_REBOOT:
		if(ctx->regs==1)
		{
			data.value=BIT_PRI_ENABLE|(mode==REGS_ENABLE_NOW?
				BITS_PRI_NOW:BITS_PRI_REBOOT);
			return ider_send(ctx,STATUS_DATA,
				(union reqdata *)&data);
		}
fail:		if(ctx->verbose)fprintf(stderr,"can't change register state\n");
		return -1;

	case REGS_DISABLE_NOW:
		if(ctx->regs!=1)goto fail;
		data.value=BIT_PRI_DISABLE|BITS_PRI_NOW;
		return ider_send(ctx,STATUS_DATA,(union reqdata *)&data);

	case REGS_DISABLE_GRACEFUL:
		if(ctx->regs!=1)goto fail;
		data.value=BIT_PRI_DISABLE|BITS_PRI_GRACE;
		return ider_send(ctx,STATUS_DATA,(union reqdata *)&data);

	case REGS_DISABLE_REBOOT:
		if(ctx->regs!=1)goto fail;
		data.value=BIT_PRI_DISABLE|BITS_PRI_REBOOT;
		return ider_send(ctx,STATUS_DATA,(union reqdata *)&data);

	default:if(ctx->verbose)fprintf(stderr,"unknown mode\n");
		return -1;
	}
}

struct iderctx *ider_init(int persist,int verbose,int debug,
	int rx_timeout,int tx_timeout,int heartbeat,int readtoc,
	int (*io)(void *parm,unsigned char *data,int len,int mode),
	int (*iowait)(void *parm,int msecs),
	void (*regstate)(void *parm,int state),void *parm)
{
	struct iderctx *ctx;

	if(rx_timeout<10000||rx_timeout>65535)
	{
		if(verbose)fprintf(stderr,"illegal rx_timeout\n");
		return NULL;
	}

	if(tx_timeout<0||tx_timeout>65535)
	{
		if(verbose)fprintf(stderr,"illegal tx_timeout\n");
		return NULL;
	}

	if(heartbeat<0||(heartbeat>0&&heartbeat<1000)||heartbeat>65535)
	{
		if(verbose)fprintf(stderr,"illegal heartbeat\n");
		return NULL;
	}

	if(!(ctx=malloc(sizeof(struct iderctx))))
	{
		if(verbose)perror("malloc");
		return NULL;
	}

	memset(ctx,0,sizeof(struct iderctx));
	ctx->persist=persist;
	ctx->verbose=verbose;
	ctx->io=io;
	ctx->iowait=iowait;
	ctx->regstate=regstate;
	ctx->parm=parm;
	ctx->debug=debug;
	ctx->session=0;
	ctx->rx_timeout=rx_timeout;
	ctx->tx_timeout=tx_timeout;
	ctx->heartbeat=heartbeat;
	ctx->readtoc=readtoc;
	ctx->sndseq=ctx->rcvseq=-1;
	ctx->scsistate=0;
	ctx->regs=-1;
	ctx->enabled=-1;

	return ctx;
}

int ider_start(struct iderctx *ctx,char *user,char *pass)
{
	struct redirinfo info;
	struct sessionparam session;

	switch(start_redirect(ctx,"IDER",&info))
	{
	case -1:if(ctx->verbose)fprintf(stderr,"connection/protocol error\n");
		return -1;

	case -2:if(ctx->verbose)fprintf(stderr,"busy, other client active, "
			"address: %s\n",inet_ntoa(info.addr));
		end_redirect(ctx);
		return -1;

	case 0:	if(ctx->verbose)fprintf(stderr,"redirect info: protocol %d.%d, "
			"firmware %d.%d, iana enterprise number %d\n",
			info.major,info.minor,info.fwmajor,info.fwminor,
			info.iana);
	}

	switch(authenticate(1,user,pass,ctx->io,ctx->parm))
	{
	case -1:if(ctx->verbose)fprintf(stderr,
			"parameter/connection/protocol error\n");
		end_redirect(ctx);
		return -1;

	case -2:if(ctx->verbose)fprintf(stderr,"authentication failure\n");
		end_redirect(ctx);
		return -1;

	case 0:	if(ctx->verbose)fprintf(stderr,"successful authentication\n");
	}

	if(ctx->verbose)fprintf(stderr,"amt ide-r connection established.\n");

	session.rx_timeout=ctx->rx_timeout;
	session.tx_timeout=ctx->tx_timeout;
	session.heartbeat=ctx->heartbeat;
	session.version=1;

	if(ider_send(ctx,OPEN_SESSION,(union reqdata *)&session))
	{
		if(ctx->verbose)fprintf(stderr,"open session request failed\n");
		end_redirect(ctx);
		return -1;
	}

	return 0;
}

int ider_end(struct iderctx *ctx)
{
	int r;

	r=end_redirect(ctx);
	free(ctx);
	return r;
}
