/*
 *  AMT IDE-R network I/O functions.
 *
 *  Copyright (C) 2014 Andreas Steinmetz <ast@domdv.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <sys/socket.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <poll.h>
#include <fcntl.h>
#include <stdio.h>
#include "auth.h"
#include "rdwr.h"

#if defined(USE_OPENSSL)
#include <openssl/ssl.h>
#include <openssl/err.h>
#elif defined(USE_GNUTLS)
#include <gnutls/gnutls.h>
#endif

union a4a6
{
	struct sockaddr_in a4;
	struct sockaddr_in6 a6;
};

#if defined(USE_OPENSSL)

static int sslinit(struct ctx *ctx)
{
	int r;
	int c=0;

	if(ctx->cacert)
	{
		SSL_load_error_strings();
		SSL_library_init();

		if(!(ctx->ctx=SSL_CTX_new(SSLv23_client_method())))
		{
			ERR_print_errors_fp(stderr);
			goto err1;
		}

#if OPENSSL_VERSION_NUMBER >= 0x1000100FL
		SSL_CTX_set_options(ctx->ctx,
			SSL_OP_NO_TLSv1_1|SSL_OP_NO_TLSv1_2);
#endif

		if(!SSL_CTX_load_verify_locations(ctx->ctx,ctx->cacert,NULL))
		{
			ERR_print_errors_fp(stderr);
			goto err2;
		}

		SSL_CTX_set_verify_depth(ctx->ctx,5);
		SSL_CTX_set_verify(ctx->ctx,SSL_VERIFY_PEER,NULL);

		if(!(ctx->ssl=SSL_new(ctx->ctx)))
		{
			ERR_print_errors_fp(stderr);
			goto err2;
		}

		if(!SSL_set_fd(ctx->ssl,ctx->fd))
		{
			ERR_print_errors_fp(stderr);
			goto err3;
		}

repeat:		if((r=SSL_connect(ctx->ssl))!=1)
		{
			switch(SSL_get_error(ctx->ssl,r))
			{
			case SSL_ERROR_WANT_READ:
			case SSL_ERROR_WANT_WRITE:
				if(++c<100)
				{
					usleep(10000);
					goto repeat;
				}
			}
			ERR_print_errors_fp(stderr);
			goto err3;
		}
	}
	else ctx->ssl=NULL;

	return 0;

err3:	SSL_free(ctx->ssl);
err2:	SSL_CTX_free(ctx->ctx);
err1:	return -1;
}

static void sslexit(struct ctx *ctx)
{
	if(ctx->ssl)
	{
		SSL_shutdown(ctx->ssl);
		SSL_free(ctx->ssl);
		SSL_CTX_free(ctx->ctx);
	}
}

static int sslpoll(SSL *ssl,struct pollfd *p,int n,int timeout)
{
	int r;

	if(ssl)if(SSL_pending(ssl))
	{
		r=poll(&p[1],n-1,0);
		if(r==-1)return -1;
		p[0].revents=POLLIN;
		return r+1;
	}
	return poll(p,n,timeout);
}

static ssize_t doread(struct ctx *ctx,void *buf,size_t count)
{
	int l;
	int c=0;

	if(!ctx->ssl)return read(ctx->fd,buf,count);
	if(!count)return 0;

repeat:	if((l=SSL_read(ctx->ssl,buf,count))>0)return l;

	switch(SSL_get_error(ctx->ssl,l))
	{
	case SSL_ERROR_WANT_READ:
	case SSL_ERROR_WANT_WRITE:
		if(++c<100)
		{
			usleep(10000);
			goto repeat;
		}
		break;
	case SSL_ERROR_WANT_X509_LOOKUP:
		return -1;
	case SSL_ERROR_ZERO_RETURN:
		return 0;
	case SSL_ERROR_SYSCALL:
		perror("SSL_read");
	case SSL_ERROR_SSL:
		ERR_print_errors_fp(stderr);
	}

	return -1;
}

static ssize_t dowrite(struct ctx *ctx,void *buf,size_t count)
{
	int l;
	int c=0;

	if(!ctx->ssl)return write(ctx->fd,buf,count);
	if(!count)return 0;

repeat:	if((l=SSL_write(ctx->ssl,buf,count))>0)return l;

	switch(SSL_get_error(ctx->ssl,l))
	{
	case SSL_ERROR_WANT_READ:
	case SSL_ERROR_WANT_WRITE:
		if(++c<100)
		{
			usleep(10000);
			goto repeat;
		}
		break;
	case SSL_ERROR_WANT_X509_LOOKUP:
		return -1;
	case SSL_ERROR_ZERO_RETURN:
		return 0;
	case SSL_ERROR_SYSCALL:
		perror("SSL_write");
	case SSL_ERROR_SSL:
		ERR_print_errors_fp(stderr);
	}

	return -1;
}

#elif defined(USE_GNUTLS)

static int vrycb(gnutls_session_t ssl)
{
	int r;
	int type;
	unsigned int status;
	gnutls_datum_t msg;

	if((r=gnutls_certificate_verify_peers3(ssl,NULL,&status))<0)
	{
		fprintf(stderr,"gnutls_certificate_verify_peers3: %s\n",
			gnutls_strerror(r));
		return GNUTLS_E_CERTIFICATE_ERROR;
	}

	if(status)
	{
		type=gnutls_certificate_type_get(ssl);
		if((r=gnutls_certificate_verification_status_print(status,type,
			&msg,0))<0)
		{
			fprintf(stderr,"gnutls_certificate_verification_"
				"status_print %s\n",gnutls_strerror(r));
		}
		else
		{
			fprintf(stderr,"certificate status: %s\n",msg.data);
			gnutls_free(msg.data);
		}
		return GNUTLS_E_CERTIFICATE_ERROR;
	}

	return 0;
}

static int sslinit(struct ctx *ctx)
{
	int r;
	const char *e;

	if(!ctx->cacert)return 0;

	if((r=gnutls_global_init()))
	{
		fprintf(stderr,"gnutls_global_init: %s\n",gnutls_strerror(r));
		goto err1;
	}

	if((r=gnutls_certificate_allocate_credentials(&ctx->cred)))
	{
		fprintf(stderr,"gnutls_certificate_allocate_credentials: "
			"%s\n",gnutls_strerror(r));
		goto err2;
	}

	if((r=gnutls_certificate_set_x509_trust_file(ctx->cred,ctx->cacert,
		GNUTLS_X509_FMT_PEM))<0)
	{
		fprintf(stderr,"gnutls_certificate_set_x509_trust_file: "
			"%s\n",gnutls_strerror(r));
		goto err3;
	}

	gnutls_certificate_set_verify_function(ctx->cred,vrycb);

	if((r=gnutls_init(&ctx->ssl,GNUTLS_CLIENT)))
	{
		fprintf(stderr,"gnutls_init: %s\n",gnutls_strerror(r));
		goto err3;
	}

	/* oh well, isn't _that_ easy ?!? :-(  ... compare to openssl ... */
	if((r=gnutls_priority_set_direct(ctx->ssl,"NONE:+AES-256-CBC:"
		"+AES-128-CBC:+3DES-CBC:+COMP-NULL:+CTYPE-X509:+VERS-SSL3.0:"
		"+SHA256:+SHA1:+RSA:%UNSAFE_RENEGOTIATION",&e)))
	{
		fprintf(stderr,"gnutls_priority_set_direct: %s\n",
			gnutls_strerror(r));
		if(r==GNUTLS_E_INVALID_REQUEST)
			fprintf(stderr,"additional info: %s\n",e);
		goto err4;
	}

	if((r=gnutls_credentials_set(ctx->ssl,GNUTLS_CRD_CERTIFICATE,
		ctx->cred)))
	{
		fprintf(stderr,"gnutls_credentials_set: %s\n",
			gnutls_strerror(r));
		goto err4;
	}

	gnutls_transport_set_int(ctx->ssl,ctx->fd);

	gnutls_handshake_set_timeout(ctx->ssl,GNUTLS_DEFAULT_HANDSHAKE_TIMEOUT);

	do
	{
		r=gnutls_handshake(ctx->ssl);
	} while(r<0&&!gnutls_error_is_fatal(r));
	if(r<0)
	{
		fprintf(stderr,"gnutls_handshake: %s\n",gnutls_strerror(r));
		goto err4;
	}

	return 0;

err4:	gnutls_deinit(ctx->ssl);
err3:	gnutls_certificate_free_credentials(ctx->cred);
err2:	gnutls_global_deinit();
err1:	return -1;
}

static void sslexit(struct ctx *ctx)
{
	if(ctx->cacert)
	{
		gnutls_deinit(ctx->ssl);
		gnutls_certificate_free_credentials(ctx->cred);
		gnutls_global_deinit();
	}
}

static int sslpoll(gnutls_session_t ssl,struct pollfd *p,int n,int timeout)
{
	int r;

	if(ssl)if(gnutls_record_check_pending(ssl))
	{
		r=poll(&p[1],n-1,0);
		if(r==-1)return -1;
		p[0].revents=POLLIN;
		return r+1;
	}
	return poll(p,n,timeout);
}

static ssize_t doread(struct ctx *ctx,void *buf,size_t count)
{
	ssize_t l;
	int c=0;
	int r;

	if(!ctx->ssl)return read(ctx->fd,buf,count);
	if(!count)return 0;

repeat:	if((l=gnutls_record_recv(ctx->ssl,buf,count))>0)return l;

	switch(l)
	{
	case GNUTLS_E_REHANDSHAKE:
		do
		{
			r=gnutls_handshake(ctx->ssl);
		} while(r<0&&!gnutls_error_is_fatal(r));
		if(r<0)
		{
			fprintf(stderr,"gnutls_handshake: %s\n",
			gnutls_strerror(r));
			return -1;
		}
	case GNUTLS_E_INTERRUPTED:
	case GNUTLS_E_AGAIN:
		if(++c<100)
		{
			usleep(10000);
			goto repeat;
		}
	default:fprintf(stderr,"gnutls_record_recv: %s\n",gnutls_strerror(l));
		return -1;
	}
}

static ssize_t dowrite(struct ctx *ctx,void *buf,size_t count)
{
	ssize_t l;
	int c=0;
	int r;

	if(!ctx->ssl)return write(ctx->fd,buf,count);
	if(!count)return 0;

repeat:	if((l=gnutls_record_send(ctx->ssl,buf,count))>0)return l;

	switch(l)
	{
	case GNUTLS_E_REHANDSHAKE:
		do
		{
			r=gnutls_handshake(ctx->ssl);
		} while(r<0&&!gnutls_error_is_fatal(r));
		if(r<0)
		{
			fprintf(stderr,"gnutls_handshake: %s\n",
			gnutls_strerror(r));
			return -1;
		}
	case GNUTLS_E_INTERRUPTED:
	case GNUTLS_E_AGAIN:
		if(++c<100)
		{
			usleep(10000);
			goto repeat;
		}
	default:fprintf(stderr,"gnutls_record_send: %s\n",gnutls_strerror(l));
		return -1;
	}
}

#else

#define sslinit(a)		0
#define sslexit(a)
#define sslpoll(a,b,c,d)	poll(b,c,d)
#define doread(a,b,c)		read(((struct ctx *)a)->fd,b,c)
#define dowrite(a,b,c)		write(((struct ctx *)a)->fd,b,c)

#endif

int iowait(void *parm,int msecs)
{
	int r=1;
	struct ctx *ctx=parm;
	struct pollfd p[2];

	if(ctx->unlock)ctx->unlock(ctx->lck);

	p[0].fd=ctx->fd;
	p[0].events=POLLIN;
	p[1].fd=ctx->efd;
	p[1].events=POLLIN;

	switch(poll(p,2,msecs))
	{
	case -1:r=-1;
		break;
	case 0:	r=0;
		break;
	default:if(p[1].revents&POLLIN)r=-1;
		else if(!(p[0].revents&POLLIN))r=0;
	}

	if(ctx->lock)ctx->lock(ctx->lck);
	return r;
}

int io(void *parm,unsigned char *data,int len,int mode)
{
	int l;
	int ll=len;
	unsigned char *pp=data;
	struct pollfd p[2];

	p[0].fd=((struct ctx *)parm)->fd;
	p[0].events=POLLIN;
	p[1].fd=((struct ctx *)parm)->efd;
	p[1].events=POLLIN;

	switch(mode)
	{
	case READ:
		for(l=0;len;data+=l,len-=l)
repeat1:		switch(sslpoll(((struct ctx *)parm)->ssl,p,2,20000))
		{
		case -1:goto repeat1;
		case 0:	return -1;
		case 1:	if(p[1].revents&POLLIN)return -1;
			if(!(p[0].revents&POLLIN))return -1;
			if((l=doread(parm,data,len))<=0)return -1;
		}
		if(((struct ctx *)parm)->iodebug)
		{
			fprintf(stderr,"<-");
			for(l=0;l<ll;l++)fprintf(stderr," %02x",pp[l]);
			fprintf(stderr,"\n");
		}
		return 0;

	case WRITE:
		if(((struct ctx *)parm)->iodebug)
		{
			fprintf(stderr,"->");
			for(l=0;l<ll;l++)fprintf(stderr," %02x",pp[l]);
			fprintf(stderr,"\n");
		}
		for(p[0].events=POLLOUT,l=0;len;data+=l,len-=l)
repeat2:		switch(poll(p,2,20000))
		{
		case -1:goto repeat2;
		case 0:	return -1;
		case 1:	if(p[1].revents&POLLIN)return -1;
			if(!(p[0].revents&POLLOUT))return -1;
			if((l=dowrite(parm,data,len))<=0)return -1;
		}
		return 0;

	case RANDOM:
		if((p[0].fd=open("/dev/urandom",O_RDONLY))==-1)return -1;
		l=read(p[0].fd,data,len);
		close(p[0].fd);
		return l==len?0:-1;
	}

	return -1;
}

static int dnslookup(char *addr,union a4a6 *res,int v4flag,int port)
{
	int r;
	struct addrinfo *ptr;
	struct addrinfo *list=NULL;
	struct addrinfo hints;
	union a4a6 a4;
	union a4a6 a6;

	memset(&a4,0,sizeof(a4));
	memset(&a6,0,sizeof(a6));

	switch(inet_pton(AF_INET,addr,&res->a4.sin_addr.s_addr))
	{
	case 1:	res->a4.sin_port=htons(port);
		res->a4.sin_family=AF_INET;
		return 0;

	case -1:perror("inet_pton");
		return -1;
	}

	switch(inet_pton(AF_INET6,addr,res->a6.sin6_addr.s6_addr))
	{
	case 1:	res->a6.sin6_port=htons(port);
		res->a6.sin6_family=AF_INET6;
		return 0;

	case -1:perror("inet_pton");
		return -1;
	}

	memset(&hints,0,sizeof(hints));
	hints.ai_family=AF_UNSPEC;
	hints.ai_flags=AI_ADDRCONFIG|AI_IDN;

	r=getaddrinfo(addr,NULL,&hints,&list);

	if(!r&&list)
	{
		memset(res,0,sizeof(union a4a6));

		for(ptr=list;ptr&&r!=3;ptr=ptr->ai_next)switch(ptr->ai_family)
		{
		case AF_INET:
			if(r&1)break;
			a4.a4.sin_family=AF_INET;
			a4.a4.sin_addr=
				((struct sockaddr_in *)ptr->ai_addr)->sin_addr;
			a4.a4.sin_port=htons(port);
			r|=1;
			break;

		case AF_INET6:
			if(r&2)break;
			a6.a6.sin6_family=AF_INET6;
			memcpy(a6.a6.sin6_addr.s6_addr,
				((struct sockaddr_in6 *)
				ptr->ai_addr)->sin6_addr.s6_addr,16);
			a6.a6.sin6_port=htons(port);
			r|=2;
			break;
		}

		freeaddrinfo(list);

		if(v4flag)
		{
			if(r&1)
			{
				res[0]=a4;
				if(r&2)res[1]=a6;
			}
			else if(r&2)res[0]=a6;
		}
		else
		{
			if(r&2)
			{
				res[0]=a6;
				if(r&1)res[1]=a4;
			}
			else if(r&1)res[0]=a4;
		}

		r=0;
	}
	else if(r)perror("getaddrinfo");

	return r;
}

int doconnect(struct ctx *ctx,char *host,int v4flag)
{
	int x;
	union a4a6 a;

	if(dnslookup(host,&a,v4flag,ctx->cacert?16995:16994))goto err2;

	if((ctx->fd=socket(a.a4.sin_family,SOCK_STREAM,IPPROTO_TCP))==-1)
	{
		perror("socket");
		goto err1;
	}

	if(connect(ctx->fd,(struct sockaddr *)&a,sizeof(a)))
	{
		perror("connect");
		goto err2;
	}

	x=1;
	if(ioctl(ctx->fd,FIONBIO,&x))
	{
		perror("ioctl");
		goto err2;
	}

	x=1;
	if(setsockopt(ctx->fd,SOL_TCP,TCP_NODELAY,&x,sizeof(x)))
	{
		perror("setsockopt");
		goto err2;
	}

	if(sslinit(ctx))goto err2;

	return 0;

err2:	close(ctx->fd);
err1:	return -1;
}

void doclose(struct ctx *ctx)
{
	sslexit(ctx);
	close(ctx->fd);
}
