/*
 *  AMT IDE-R command line utility.
 *
 *  Copyright (C) 2014 Andreas Steinmetz <ast@domdv.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <pthread.h>
#include <sys/eventfd.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <poll.h>
#include <fcntl.h>
#include <stdio.h>
#include "ider.h"
#include "rdwr.h"

static pthread_mutex_t mtx=PTHREAD_MUTEX_INITIALIZER;
static int diskfd=-1;
static int cddvdfd=-1;
static int quiet=0;
static int msg[2];

static void *processor(void *data)
{
	struct iderctx *ctx=data;
	sigset_t set;

	sigfillset(&set);
	pthread_sigmask(SIG_BLOCK,&set,NULL);

	pthread_mutex_lock(&mtx);

	while(1)if(ider_engine(ctx))break;

	pthread_mutex_unlock(&mtx);

	quiet=1;
	kill(getpid(),SIGHUP);

	pthread_exit(NULL);
}

static void handler(int unused)
{
	if(!quiet)printf("\n");
}

static void usage(void)
{
  fprintf(stderr,"Usage: idertool <options> host\n"
  "Options are:\n"
  "-d                         increase debug mode\n"
  "-v                         verbose mode\n"
  "-4                         prefer IPv4 on dns lookups\n"
  "-P                         request persistent connection\n"
#if defined(USE_OPENSSL) || defined(USE_GNUTLS)
  "-c cacert                  use the provided cacert file and enable ssl\n"
#endif
  "-u user                    amt user ($IDER_USER, default is admin)\n"
  "-p password                amt password ($IDER_PASS)\n"
  "-r timeout                 host receive timeout (10-65s, default 10)\n"
  "-t timeout                 host transmit timeout (0-65s, default 0, 0=off)\n"
  "-h timeout                 heartbeat timeout (0-65s, default 5, 0=off)\n"
  "-T                         disable SCSI READ_TOC command (testing)\n"
  "-S                         show register status\n"
  "-E now|reboot              enable registers now/on reboot\n"
  "-D now|graceful|reboot     disable registers now/graceful/on reboot\n"
  "-f image                   attach disk image read/write\n"
  "-F image                   attach disk image read only\n"
  "-C image                   attach cd/dvd data image read only\n");
  exit(1);
}

static void regstate(void *unused,int state)
{
	unsigned char val;

	val=state&0xff;
	state=write(msg[1],&val,1);
}

int main(int argc,char *argv[])
{
	int x;
	int r=1;
	int state=0;
	int status=0;
	int mode=0;
	int debug=0;
	int verbose=0;
	int v4flag=0;
	int persist=0;
	int enable=0;
	int disable=0;
	int rx_timeout=10;
	int tx_timeout=0;
	int heartbeat=5;
	int readtoc=1;
	unsigned char val;
	char *addr;
	char *user;
	char *pass;
	char *disk=NULL;
	char *cddvd=NULL;
	struct iderctx *ctx;
	uint64_t dummy=1;
	struct pollfd p;
	struct ctx c;
	pthread_t h;
	sigset_t set;

	c.iodebug=0;
	c.cacert=NULL;
	c.lck=&mtx;
	c.lock=(void (*)(void *))pthread_mutex_lock;
	c.unlock=(void (*)(void *))pthread_mutex_unlock;

	if(!(user=getenv("IDER_USER")))user="admin";
	pass=getenv("IDER_PASS");

	while((x=getopt(argc,argv,"dv4u:p:PE:D:f:F:C:r:t:h:Sc:T"))!=-1)switch(x)
	{
	case 'd':
		debug++;
		break;

	case 'v':
		verbose=1;
		break;

	case '4':
		v4flag=1;
		break;

	case 'u':
		user=optarg;
		break;

	case 'p':
		pass=optarg;
		break;

	case 'P':
		persist=1;
		break;

	case 'E':
		if(!strcasecmp(optarg,"now"))enable=1;
		else if(!strcasecmp(optarg,"reboot"))enable=2;
		else usage();
		break;

	case 'D':
		if(!strcasecmp(optarg,"now"))disable=1;
		else if(!strcasecmp(optarg,"reboot"))disable=2;
		else if(!strcasecmp(optarg,"graceful"))disable=3;
		else usage();
		break;

	case 'f':
		mode=DEV_RW;
		disk=optarg;
		break;

	case 'F':
		mode=DEV_RO;
		disk=optarg;
		break;

	case 'C':
		mode=DEV_RO;
		cddvd=optarg;
		break;

	case 'r':
		rx_timeout=atoi(optarg);
		break;

	case 't':
		tx_timeout=atoi(optarg);
		break;

	case 'h':
		heartbeat=atoi(optarg);
		break;

	case 'S':
		status=1;
		break;

	case 'T':
		readtoc=0;
		break;

#if defined(USE_OPENSSL) || defined(USE_GNUTLS)
	case 'c':
		c.cacert=optarg;
		break;
#endif
	default:usage();
	}

	rx_timeout*=1000;
	tx_timeout*=1000;
	heartbeat*=1000;

	if(optind!=argc-1)usage();
	else addr=argv[optind];

	if(enable&&disable)usage();

	if(!pass)usage();

	if(debug>1)c.iodebug=1;

	if((c.efd=eventfd(0,0))==-1)
	{
		perror("eventfd");
		goto err1;
	}

	signal(SIGINT,handler);
	signal(SIGHUP,handler);
	signal(SIGTERM,handler);
	signal(SIGQUIT,handler);
	signal(SIGPIPE,SIG_IGN);

	if(doconnect(&c,addr,v4flag))goto err2;

	if(!(ctx=ider_init(persist,verbose,debug,rx_timeout,tx_timeout,
		heartbeat,readtoc,io,iowait,regstate,&c)))goto err3;

	if(ider_start(ctx,user,pass))goto err4;

	if(pipe(msg))
	{
		perror("pipe");
		goto err4;
	}

	p.fd=msg[0];
	p.events=POLLIN;

	if(pthread_create(&h,NULL,processor,ctx))
	{
		perror("pthread_create");
		goto err5;
	}

	if(enable||disable||status)switch(poll(&p,1,3000))
	{
	case -1:perror("poll");
		goto err6;

	case 1:	if(p.revents&POLLIN)
		{
			state=read(msg[0],&val,1);
			if(state==1)if((state=val)!=REGS_NONE)break;
		}
	case 0:	enable=disable=status=0;
		if(verbose)fprintf(stderr,"can't get register state\n");
		break;
	}

	if(enable&&state==REGS_ENABLED)enable=0;
	if(disable&&state==REGS_DISABLED)disable=0;

	if(status)switch(state)
	{
	case REGS_ENABLED:
		printf("enabled\n");
		break;

	case REGS_DISABLED:
		printf("disabled\n");
		break;
	}

	switch(enable)
	{
	case 1:	pthread_mutex_lock(&mtx);
		state=ider_regs(ctx,REGS_ENABLE_NOW);
		pthread_mutex_unlock(&mtx);
		break;

	case 2:	pthread_mutex_lock(&mtx);
		state=ider_regs(ctx,REGS_ENABLE_REBOOT);
		pthread_mutex_unlock(&mtx);
		break;
	}

	switch(disable)
	{
	case 1:	pthread_mutex_lock(&mtx);
		state=ider_regs(ctx,REGS_DISABLE_NOW);
		pthread_mutex_unlock(&mtx);
		break;

	case 2:	pthread_mutex_lock(&mtx);
		state=ider_regs(ctx,REGS_DISABLE_REBOOT);
		pthread_mutex_unlock(&mtx);
		break;

	case 3:	pthread_mutex_lock(&mtx);
		state=ider_regs(ctx,REGS_DISABLE_GRACEFUL);
		pthread_mutex_unlock(&mtx);
		break;
	}

	if((enable||disable)&&state)
	{
		if(verbose)fprintf(stderr,"can't set register state\n");
		goto err6;
	}

	if((enable||disable)&&!state)switch(poll(&p,1,3000))
	{
	case -1:perror("poll");
		goto err6;

	case 1:	if(p.revents&POLLIN)
		{
			state=read(msg[0],&val,1);
			if(state==1)if((state=val)!=REGS_FAIL)break;
		}
	case 0:	if(verbose)fprintf(stderr,"can't set register state\n");
		goto err6;
	}

	if(disk)
	{
		if((diskfd=open(disk,mode==DEV_RO?O_RDONLY:O_RDWR))==-1)
		{
			perror("open");
			goto err6;
		}

		pthread_mutex_lock(&mtx);
		state=ider_device(ctx,DEV_FLOPPY,mode,diskfd);
		pthread_mutex_unlock(&mtx);

		if(state)
		{
			if(verbose)fprintf(stderr,"can't attach disk image\n");
			goto err6;
		}
	}

	if(cddvd)
	{
		if((cddvdfd=open(cddvd,O_RDONLY))==-1)
		{
			perror("open");
			goto err6;
		}

		pthread_mutex_lock(&mtx);
		state=ider_device(ctx,DEV_CDDVD,DEV_RO,cddvdfd);
		pthread_mutex_unlock(&mtx);

		if(state)
		{
			if(verbose)fprintf(stderr,"can't attach cd/dvd "
				"image\n");
			goto err6;
		}
	}

	if(disk||cddvd)
	{
		sigemptyset(&set);
		sigsuspend(&set);
	}

	r=0;

err6:	if(cddvdfd!=-1)
	{
		pthread_mutex_lock(&mtx);
		state=ider_device(ctx,DEV_CDDVD,DEV_NONE,-1);
		pthread_mutex_unlock(&mtx);
		close(cddvdfd);
	}

	if(diskfd!=-1)
	{
		pthread_mutex_lock(&mtx);
		state=ider_device(ctx,DEV_FLOPPY,DEV_NONE,-1);
		pthread_mutex_unlock(&mtx);
		close(diskfd);
	}

	dummy=write(c.efd,&dummy,sizeof(dummy));

	pthread_join(h,NULL);

	dummy=read(c.efd,&dummy,sizeof(dummy));

err5:	close(msg[0]);
	close(msg[1]);
err4:	ider_end(ctx);
err3:	doclose(&c);
err2:	close(c.efd);
err1:	return r;
}
